// vim:set tabstop=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include <vector>
#include <string>
#include "geometry.hpp"
#include "uuc3d.hpp"

//! \file
//! \brief Interpolation functions for 3d markers.

using std::vector;
using std::string;
using UuIcsC3d::FrameData;
using UuIcsC3d::DataPoint3d;

using namespace UuIcsGeom;

namespace {

	///Range to look for valid marker before and after a point.
	/// If no valid point is found within this range, no interpolation will be performed
	const int SrchRg = 400; // was: 25; 

	enum GenerationMethod { NoMethod=0, LinearInterpolation=1, VirtualMarker=2, XhInterpolation2=3, XinghanInterpolation=5};
	struct IntRange {
		int begin;
		int end;
		IntRange() {}
		IntRange(int m_begin, int m_end): begin(m_begin), end(m_end) {}
	};

	// estimate the position of invalid markers
	// via interpolating between its previous & next valid locations 
	IntRange interpolate_marker(
		vector< vector<Point3d> > &pos,
		vector< vector<bool> > const &valid,
		int f, int m)
	{
		int const frameNr = pos.size();
		assert(valid.size()==pos.size());
		assert (0<=f);
		assert(f<frameNr);
		int const markerNo=pos[0].size();
		assert(0<=m);
		assert(m<markerNo);
		assert(!valid[f][m]);
		// Find the previous valid position (or from-1)
		int from = f-SrchRg;
		if (from<0) from = 0;
		int prevFIdx = f-1;
		while(((prevFIdx>=from) && !valid[prevFIdx][m]))
			prevFIdx--;
		// Find the next valid position (or 'to')
		int to = f+SrchRg;
		if (to>frameNr) to=frameNr;
		int forwFIdx = f+1;
		while((forwFIdx<to)&&(!valid[forwFIdx][m]))
			forwFIdx++;
		// Now interpolate or extrapolate the marker
		if (prevFIdx<from) {
			if(to<=forwFIdx) {
				// No valid marker found
				return IntRange(f, f);
			} else {
				Point3d fill_point = pos[forwFIdx][m];
				for (int cf=prevFIdx+1; cf<forwFIdx; ++cf) {
					pos[cf][m] = fill_point;
				}
			}
		} else {
			if(to<=forwFIdx) {
				Point3d fill_point = pos[prevFIdx][m];
				for (int cf=prevFIdx+1; cf<forwFIdx; ++cf) {
					pos[cf][m] = fill_point;
				}
			} else {
				Vector3d inTTrans = pos[forwFIdx][m] - pos[prevFIdx][m];
				Point3d start_point = pos[prevFIdx][m];
				for (int cf=prevFIdx+1; cf<forwFIdx; ++cf) {
					double inTRatio = 1.0*(cf - prevFIdx)/(forwFIdx - prevFIdx);
					pos[cf][m] = Point3d(start_point + inTRatio*inTTrans);
				}
			}
		}
		return IntRange(prevFIdx+1,forwFIdx);
	}

	IntRange estimate_marker_positions(
		vector< vector<Point3d> > &pos,  
		vector< vector<bool> > const &valid,
		int f, int m)
//	estimate_marker_positions estimates the position of a marker (m) around a requested
//	frame (f). It takes into account the position changes of the other markers.
//
//	pos[fi][mi] is the position of marker mi in frame fi.
//	valid[fi][mi] indicates whether pos[fi][mi] is valid.
//	The [begin, end) range of the return value indicates which frames were updated.
	{
		int const frameNr = pos.size();
		assert(valid.size()==pos.size());
		assert (0<=f);
		assert(f<frameNr);
		int const markerNo=pos[0].size();
		assert(0<=m);
		assert(m<markerNo);
		assert(!valid[f][m]);
		// Find the previous valid position (or from-1)
		int from = f-SrchRg;
		if (from<0) from = 0;
		int prevFIdx = f-1;
		while(((prevFIdx>=from) && !valid[prevFIdx][m]))
			prevFIdx--;
		// Find the next valid position (or 'to')
		int to = f+SrchRg;
		if (to>frameNr) to=frameNr;
		int forwFIdx = f+1;
		while((forwFIdx<to)&&(!valid[forwFIdx][m]))
			forwFIdx++;

		if (prevFIdx<from && to<=forwFIdx) {
			// No valid marker is available in any frame
			// Return empty range
			return IntRange(f, f);
		}
		int missingFrameCount = forwFIdx-prevFIdx-1;
		vector<Point3d> bwPoints(missingFrameCount), fwPoints(missingFrameCount);

		if(prevFIdx>=from) {
			// Compute forward extrapolation from previous valid position
			Point3d *prev_pt = &pos[prevFIdx][m];
			for(int rf = 0; rf!=missingFrameCount; ++rf ) {
				int fS = prevFIdx+rf+1;
				// get the average translation based on the other valid markers
				int mCount = 0; // number of markers involved in the average
				Vector3d transl(0,0,0);
				for(int mm = 0; mm < markerNo; mm++ ) {			
					if((mm != m)&&valid[fS][mm] && valid[fS-1][mm]) {
						transl += (pos[fS][mm]-pos[fS-1][mm]);                                       
						mCount++;
					}
				}
				if (mCount >0)
					transl *= 1.0/mCount;
				bwPoints[rf] = Point3d((*prev_pt)+transl);
				prev_pt = &bwPoints[rf];
			}
		}

		if(forwFIdx<to) {
			// Compute backward extrapolation from next valid position
			Point3d *prev_pt = &pos[forwFIdx][m];
			for(int rf = missingFrameCount; rf!=0; --rf ) {
				int fS = prevFIdx+rf;
				// get the average translation based on the other valid markers
				int mCount = 0; // number of markers involved in the average
				Vector3d transl(0,0,0);
				for(int mm = 0; mm < markerNo; mm++ ) {			
					if((mm != m)&&valid[fS][mm] && valid[fS+1][mm]) {
						transl += (pos[fS][mm]-pos[fS+1][mm]);                                       
						mCount++;
					}
				}
				if (mCount >0)
					transl *= 1.0/mCount;
				fwPoints[rf-1] = Point3d((*prev_pt)+transl);
				prev_pt = &fwPoints[rf-1];
			}
		}
		if(prevFIdx>=from) {
			if (forwFIdx<to) {
				// linear interpolation of forward and backward estimate
				for(int rf = 0; rf!=missingFrameCount; ++rf ) {
					int fS = prevFIdx+rf+1;
					pos[fS][m] = Point3d(
						  1.0*(fS-prevFIdx)/(missingFrameCount+1)*fwPoints[rf]
						+ 1.0*(forwFIdx-fS)/(missingFrameCount+1)*bwPoints[rf]);
				}
			} else {
				for(int rf = 0; rf!=missingFrameCount; ++rf ) {
					int fS = prevFIdx+rf+1;
					pos[fS][m] =bwPoints[rf];
				}
			}
		} else {
			for(int rf = 0; rf!=missingFrameCount; ++rf ) {
				int fS = prevFIdx+rf+1;
				pos[fS][m] =fwPoints[rf];
			}
		}
		return IntRange(prevFIdx+1, forwFIdx);
	}
}


//! \brief Fill in missing positions for markers in a group, Xinghans method
void interpolate_group1(vector<FrameData> &frame_data, vector<int> const &point_index)
{
	size_t const frameNr = frame_data.size();
	size_t const pointNr = point_index.size();
	if (pointNr==0)
		return;
	// store the marker positions and validity in a working copy 
	vector< vector<Point3d> > pos(frameNr, vector<Point3d>(pointNr, Point3d(0,0,0)));
	vector< vector<bool> > valid(frameNr, vector<bool>(pointNr, false));
	for( size_t f = 0; f != frameNr; ++f ) {
		FrameData &cur_frame = frame_data[f];
		for(size_t p= 0; p < pointNr; ++p ) {
			int cur_index = point_index[p];
			if (cur_index<0)
				continue;
			DataPoint3d const &dp = cur_frame.points[cur_index];
			if(dp.is_valid()) {			
				pos[f][p] = Point3d(dp.coordinates());
				valid[f][p] = true;
			}
		}		
	}

	// estimate the invalid markers by valid markers in the same group and
	// its coordinates in previous next frames
	if (pointNr == 1) {
		for(size_t f = 0; f != frameNr; f++ ) {
			if (!valid[f][0]) {
				IntRange range = interpolate_marker(pos, valid, f, 0);
				for (int i=range.begin; i!=range.end; ++i) {
					valid[i][0] = true;
					frame_data[i].points[point_index[0]].set_value(
						pos[i][0].coords, LinearInterpolation);
				}
			}
		}
	} else {
		vector< vector<bool> > ori_valid(valid);
		for(size_t f = 0; f != frameNr; f++ ) {
			for(size_t m = 0; m != pointNr; m++ ) {
				if(!valid[f][m]) {
					IntRange range = estimate_marker_positions(
									pos, ori_valid, f, m);
					for (int i=range.begin; i!=range.end; ++i) {
						valid[i][m] = true;
						frame_data[i].points[point_index[m]].set_value(
								pos[i][m].coords, XinghanInterpolation);
					}
				}
			}
		}
	}
}
