// vim:set tabstop=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include "skeleton_info.hpp"
#include "uuc3d.hpp"
#include <iostream>
#include <set>
#include "subjects.hpp"
#include "label_anonymous.hpp"

using std::string;
using std::vector;
using std::cerr;
using UuIcsC3d::SpacePaddedString;
using UuIcsC3d::FrameData;
using UuIcsC3d::LabeledSkeleton;
using UuIcsC3d::LabeledStretch;

UuIcsC3d::SkeletonInfo const &get_uu_skeleton_info();

void interpolate_group1(vector<FrameData> &frame_data,
		vector<int> const &point_index);

void interpolate_positions1(vector<FrameData> &frame_data,
		vector<LabeledSkeleton> const &subjects)
{
	for (size_t s=0; s!= subjects.size(); ++s) {
		vector< vector<int> > const &point_index = subjects[s].point_index();
		for (size_t g=0; g!= point_index.size(); ++g) {
			interpolate_group1(frame_data, point_index[g]);
		}
	}
}

void remove_uncertain(vector<FrameData> &frame_data)
{
    size_t const fc = frame_data.size();
	for (size_t f=0; f!= fc; ++f) {
		vector<UuIcsC3d::DataPoint3d> &points = frame_data[f].points;
		for (size_t l=0; l != points.size(); ++l) {
			if (points[l].status() == UuIcsC3d::DataPoint3d::Generated) {
					points[l].invalidate();
			}
		}
	}
}

bool no_generated_points(vector<FrameData> const&frame_data)
{
    size_t const fc = frame_data.size();
	for (size_t f=0; f!= fc; ++f) {
		vector<UuIcsC3d::DataPoint3d> const&points = frame_data[f].points;
		for (size_t l=0; l != points.size(); ++l) {
			if (points[l].status() == UuIcsC3d::DataPoint3d::Generated) {
				std::cerr<<"Marker "<<l<<" frame "<<f<<"\n";
				return false;
			}
		}
	}
	return true;
}

bool no_generated_points_for_marker(vector<FrameData> const&frame_data, int m)
{
	size_t const fc = frame_data.size();
	for (size_t f=0; f!= fc; ++f) {
		//vector<UuIcsC3d::DataPoint3d> const&points = frame_data[f].points;
		if (frame_data[f].points[m].status() == UuIcsC3d::DataPoint3d::Generated) {
			std::cerr<<"Marker "<<m<<" frame "<<f<<"\n";
			return false;
		}
	}
	return true;
}

void do_work(string infile, string outfile)
{
    UuIcsC3d::C3dFileInfo fi(infile);
    int ppf = fi.points_per_frame(); 
    if (ppf <= 0)
        return;
    int fc = fi.frame_count();
    vector<SpacePaddedString> labels(fi.point_labels());
    vector<LabeledSkeleton> subjects;
    vector<unsigned> no_label = fill_skeletons(
			subjects, labels, get_uu_skeleton_info());
    for (int i=labels.size(); i<ppf; ++i) {
        labels.push_back(SpacePaddedString("<no label>"));
    }
    std::auto_ptr<UuIcsC3d::C3dFile> filep(fi.open());
    vector<FrameData> frame_data(fc);
    for (int i=0; i<fc; ++i) {
        filep->get_frame_data(frame_data[i], i);
    }
    filep.reset();

	double const MaxSpeed = 20.0;
	while (true) {
		assert(no_generated_points(frame_data));
		interpolate_positions1(frame_data, subjects);
		vector<LabeledStretch> relabeled = label_anonymous_markers(
			frame_data, subjects, no_label, MaxSpeed, labels);
		if (relabeled.empty())
			break;
		remove_uncertain(frame_data);
	}
	remove_uncertain(frame_data);
	assert(no_generated_points(frame_data));
    if (!write_with_same_header(outfile, fi, frame_data))
        cerr<<"Error while writing file "<<outfile<<".\n";
}


char *progname=0;

void print_usage(std::ostream &os)
{
    os<<"Usage:\n";
    os<<progname<<" input_c3d output_c3d\n";
}

int main(int argc, char* argv[])
{
    progname = argv[0];
    if (argc<3) {
        print_usage(cerr);
        return 1;
    }
    try {
        do_work(argv[1], argv[2]);
    } catch (UuIcsC3d::OpenError const &err) {
        cerr<<"An error occurred while opening file "<<err.filename();
        return 2;
	}
    return 0;
}
