// vim:set tabstop=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include "skeleton_info.hpp"
#include <cassert>
#include <algorithm>
#include <stdlib.h>

using std::string;
using std::vector;

using UuIcsC3d::ConstNameRange;

int UuIcsC3d::MarkerNames::count() const
{
	return std::distance(begin(), end());
}

namespace {

	void check_precondition(bool val)
	{
		if (!val)
			abort();
	}

	template <class T, int N>
	int array_size(T (&)[N]) { return N;}

	template <class T, int N>
	T* array_beyond(T (&arr)[N]) { return arr+N;}

	string uu_skeleton_names[] = {string("lankfr"), string("lankbk"), string("lankis"),
		string("lkneefr"), string("lknssbk"), string("lkneeis"),
		string("ltopleg"),
		string("rankfr"), string("rankbk"), string("rankis"),
		string("rkneefr"), string("rkneebk"), string("rkneeis"),
		string("rtopleg"),
		string("lasi"), string("rasi"), string("lpsi"), string("rpsi"),
		string("fneck"), string("bneck"),
		string("fhead"), string("lhead"), string("rhead"),
		string("lshld"),
		string("lelbtop"), string("lelblow"), string("lelbext"),
		string("lwrtop"), string("lwrlow"), string("lwrext"),
		string("rshld"),
		string("relbtop"), string("relblow"), string("relbext"),
		string("rwrtop"), string("rwrlow"), string("rwrext")};
	int uu_skeleton_beyond_count[] = {3,6,7,10,13,14,18,20,23,24,27,30,31,34,37};

	template <class Type, int Dim>
	inline void add_nums(vector< int > &result, Type (&arr)[Dim])
	{
		result.insert(result.end(),arr+0, arr+Dim);
	}

	int const GroupCount = 15;
	string group_names[GroupCount] = {"lankle", "lknee", "lleg", "rankle",
        "rknee", "rleg", "pelvis", "neck", "head", "lshld", "lelbow",
        "lwrist", "rshld", "relbow", "rwrist"};


	vector< vector <int> > const &the_group_neighbors()
	{
		static vector< vector<int> > result;
		static bool ready=false;
		if (!ready) {
			ready=true;
			result.reserve(GroupCount);
			vector<int> empty;

			result.push_back(empty);
			int lankle[1] = {1};
			add_nums(result.back(), lankle);
			result.push_back(empty);
			int lknee[2] = {0,2};
			add_nums(result.back(), lknee);
			result.push_back(empty);
			int lleg[2] = {1,6};
			add_nums(result.back(), lleg);
			result.push_back(empty);
			int rankle[1] = {4};
			add_nums(result.back(), rankle);
			result.push_back(empty);
			int rknee[2] = {3,5};
			add_nums(result.back(), rknee);
			result.push_back(empty);
			int rleg[2] = {4,6};
			add_nums(result.back(), rleg);
			result.push_back(empty);
			int pelvis[3] = {2,5,7};
			add_nums(result.back(), pelvis);
			result.push_back(empty);
			int neck[2] = {6,8};
			add_nums(result.back(), neck);
			result.push_back(empty);
			int head[1] = {7};
			add_nums(result.back(), head);
			result.push_back(empty);
			int lshld[1] = {10};
			add_nums(result.back(), lshld);
			result.push_back(empty);
			int lelbow[2] = {9,11};
			add_nums(result.back(), lelbow);
			result.push_back(empty);
			int lwrist[1] = {10};
			add_nums(result.back(), lwrist);
			result.push_back(empty);
			int rshld[1] = {13};
			add_nums(result.back(), rshld);
			result.push_back(empty);
			int relbow[2] = {12,14};
			add_nums(result.back(), relbow);
			result.push_back(empty);
			int rwrist[1] = {13};
			add_nums(result.back(), rwrist);
		}
		assert(result.size()==static_cast<size_t>(GroupCount));
		return result;
	}
}

//! \brief Class that supplies information about the skeleton used in the video
//! benchmark by Utrecht University in 2010.
struct UUSkeletonInfo: UuIcsC3d::SkeletonInfo {
	int count() const;
	std::string const *begin() const;
	std::string const *end() const;

	int group_count() const;
	string group_name(int i) const;
	ConstNameRange group_members(int i) const;
	vector<int> group_connections(int group) const;
	GrEl to_grel(int label_no) const;
	int to_label_no(GrEl const & grel) const;
};

int UUSkeletonInfo::count() const
{
	return array_size(uu_skeleton_names);
}

std::string const *UUSkeletonInfo::begin() const
{
	return uu_skeleton_names;
}

std::string const *UUSkeletonInfo::end() const
{
	return array_beyond(uu_skeleton_names);
}

int UUSkeletonInfo::group_count() const
{
	assert(GroupCount==array_size(uu_skeleton_beyond_count));
	return GroupCount;
}


string UUSkeletonInfo::group_name(int group) const
{
	check_precondition(0<=group && group<GroupCount);
	return group_names[group];
}

ConstNameRange UUSkeletonInfo::group_members(int group) const
{
	check_precondition(0<=group && group<GroupCount);
	int base = group==0 ? 0 : uu_skeleton_beyond_count[group-1];	
	return ConstNameRange(uu_skeleton_names+base, 
		uu_skeleton_names+uu_skeleton_beyond_count[group]);
}

vector<int> UUSkeletonInfo::group_connections(int group) const
{
	check_precondition(0<=group && group<GroupCount);
	return the_group_neighbors()[group];
}

UUSkeletonInfo::GrEl UUSkeletonInfo::to_grel(int label_no) const
{
	assert(label_no>=0);
	int *abp = array_beyond(uu_skeleton_beyond_count);
	int *lbp = std::lower_bound(uu_skeleton_beyond_count, abp, label_no+1);
	assert(lbp != abp);
	GrEl result;
	result.gr = std::distance(uu_skeleton_beyond_count, lbp);
	result.el = result.gr==0? label_no 
		: label_no-uu_skeleton_beyond_count[result.gr-1];
	return result;
}

int UUSkeletonInfo::to_label_no(GrEl const & grel) const
{
	assert(grel.gr < array_size(uu_skeleton_beyond_count));
	if (grel.gr < array_size(uu_skeleton_beyond_count))
		return -1;
	int result= grel.gr==0? grel.el : uu_skeleton_beyond_count[grel.gr-1]+grel.el;
	assert(result<uu_skeleton_beyond_count[grel.gr]);
	if (result>=uu_skeleton_beyond_count[grel.gr])
		return -1;
	return result;
}

UuIcsC3d::SkeletonInfo const &get_uu_skeleton_info()
{
	static UUSkeletonInfo the_uu_skeleton;
	return the_uu_skeleton;
}


namespace {
	struct UUVirtualMarkerInfo: UuIcsC3d::MarkerNames {
		int count() const;
		std::string const *begin() const;
		std::string const *end() const;
	};


	string VMLabels[] = { string("pelvis"), string("thorax"), string("head"), 
		string("rshoulder"), string("relbow"), string("rwrist"), 
		string("lshoulder"), string("lelbow"), string("lwrist"), 
		string("rhip"), string("rknee"), string("rankle"), 
		string("lhip"), string("lknee"), string("lankle") };


	int UUVirtualMarkerInfo::count() const
	{
		return array_size(VMLabels);
	}
	std::string const *UUVirtualMarkerInfo::begin() const
	{
		return VMLabels;
	}

	std::string const *UUVirtualMarkerInfo::end() const
	{
		return array_beyond(VMLabels);
	}
}


UuIcsC3d::MarkerNames const &get_uu_virtual_marker_info()
{
	static UUVirtualMarkerInfo the_uu_skeleton;
	return the_uu_skeleton;
}


