// vim:set tabstop=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include "subjects.hpp"
#include <map>

using std::vector;
using std::string;

UuIcsC3d::LabeledSkeleton::LabeledSkeleton(string const &name,
        SkeletonInfo const &skel_info)
: m_name(name), m_label_count(0), m_point_index(skel_info.group_count())
{
	for (int i=0; i<skel_info.group_count(); ++i) {
		ConstNameRange names = skel_info.group_members(i);
		size_t group_size = std::distance(names.begin(), names.end());
		m_label_count += group_size;
		m_point_index[i].resize(group_size, -1);
	}
}

UuIcsC3d::LabeledSubject::LabeledSubject(
				string const &name, MarkerNames const &marker_names)
: m_name(name), m_label_count(marker_names.count()),
		m_point_index(marker_names.count(), -1)
{
}

namespace {
	struct SameNamePred {
		SameNamePred(string const &name):m_name(name) {}
		bool operator()(UuIcsC3d::LabeledSkeleton const &ls)
        {return ls.name()==m_name;}
		bool operator()(UuIcsC3d::LabeledSubject const &ls)
        {return ls.name()==m_name;}
	private:
		string m_name;
	};

	struct GroupElement {
		int group;
		int element;
	};
}

vector<unsigned> UuIcsC3d::fill_skeletons(
        vector<LabeledSkeleton> &subjects,
        vector<SpacePaddedString> const &labels,
        SkeletonInfo const &skel_info)
// Parse the labels and construct the LabeledSubjects.
{
	using std::find;
//	SkeletonInfo const &skel_info(get_uu_skeleton_info());
	LabeledSkeleton blank_subject("blank",skel_info);
	typedef std::map<string, GroupElement> MarkerMap;
	MarkerMap marker_id;
	// fill the map with the marker names
	for (int gr=0; gr<skel_info.group_count(); ++gr) {
		GroupElement grel;
		grel.group=gr;
		ConstNameRange names = skel_info.group_members(gr);
		string const *cur;
		for (cur=names.begin(), grel.element=0; cur!=names.end();
                ++cur, ++grel.element) {
			marker_id[*cur]=grel;
		}
	}
	vector<unsigned> no_label;
	subjects.clear();
	for (unsigned lab_id=0; lab_id<labels.size(); ++lab_id) {
		// split the label in a subject and a marker part
		string cur_label = labels[lab_id].stripped();
		string::size_type pos = cur_label.find_last_of(':');
		string subject_id;
		string marker_name;
		if (pos==string::npos) {
			marker_name = cur_label;
		} else {
			subject_id=cur_label.substr(0, pos);
			marker_name=cur_label.substr(pos+1);
		}
		// Look up group and element belonging to marker_name. 
		// Then set the index of the element found to the current label index.
		MarkerMap::iterator marker_pos = marker_id.find(marker_name);
		if (marker_pos == marker_id.end()) {
			no_label.push_back(lab_id);
			continue;
		}
		GroupElement grel = marker_pos->second;
		// Find existing subject with current subject name
        //  or insert a new subject with that name
		vector<LabeledSkeleton>::iterator ls_it = find_if(
                subjects.begin(), subjects.end(), SameNamePred(subject_id));
		if (ls_it==subjects.end()) {
			ls_it=subjects.insert(subjects.end(),blank_subject);
			ls_it->set_name(subject_id);
		}
		ls_it->set_index(grel.group,grel.element,lab_id);
	}
	return no_label;
}

namespace {
	struct NameIndex {
		NameIndex(string const &name_in, int index_in): name(name_in), index(index_in) {}
		string name;
		int index;
	};

	struct MiLessByName {
		inline bool operator()(NameIndex const &ni1, NameIndex const &ni2)
		{
			return ni1.name<ni2.name;
		}
		inline bool operator()(string const &name, NameIndex const &ni2)
		{
			return name<ni2.name;
		}
		inline bool operator()(NameIndex const &ni1, string const &name)
		{
			return ni1.name<name;
		}
	};
}

vector<unsigned> UuIcsC3d::fill_subjects(
        vector<LabeledSubject> &subjects,
        vector<SpacePaddedString> const &labels,
        MarkerNames const &marker_names)
// Parse the labels and construct the LabeledSubjects.
{
	using std::find;
	LabeledSubject blank_subject("blank",marker_names);
//	typedef std::map<string, GroupElement> MarkerMap;
	vector<NameIndex> marker_id;
	// fill the map with the marker names
	{
		string const *cur;
		int idx=0;
		for (cur=marker_names.begin(), idx=0; cur!=marker_names.end();
                ++cur, ++idx) {
			marker_id.push_back(NameIndex(*cur, idx));
		}
		sort(marker_id.begin(), marker_id.end(), MiLessByName());
	}
	vector<unsigned> no_label;
	subjects.clear();
	for (unsigned lab_id=0; lab_id<labels.size(); ++lab_id) {
		// split the label in a subject and a marker part
		string cur_label = labels[lab_id].stripped();
		string::size_type pos = cur_label.find_last_of(':');
		string subject_id;
		string marker_name;
		if (pos==string::npos) {
			marker_name = cur_label;
		} else {
			subject_id=cur_label.substr(0, pos);
			marker_name=cur_label.substr(pos+1);
		}
		// Look up group and element belonging to marker_name. 
		// Then set the index of the element found to the current label index.
		vector<NameIndex>::iterator marker_pos = lower_bound(marker_id.begin(), marker_id.end(), marker_name, MiLessByName());
		if (marker_pos->name != marker_name) {
			no_label.push_back(lab_id);
			continue;
		}
		int idx = marker_pos->index;
		// Find existing subject with current subject name
        //  or insert a new subject with that name
		vector<LabeledSubject>::iterator ls_it = find_if(
                subjects.begin(), subjects.end(), SameNamePred(subject_id));
		if (ls_it==subjects.end()) {
			ls_it=subjects.insert(subjects.end(),blank_subject);
			ls_it->set_name(subject_id);
		}
		ls_it->set_index(idx,lab_id);
	}
	return no_label;
}
